/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	RBD/NJN
 *
 *	$Id: lamunpack.c,v 6.1 96/11/23 22:53:20 nevin Rel $
 *
 *	Function:	- MPI data unpacking functions
 *			- handles recursive data types
 */

#include <mpi.h>
#include <mpisys.h>
#include <mpi_types.h>
#include <portable.h>
#include <t_types.h>
#include <typical.h>

/*
 * local functions
 */
static int		unpack_hvector();
static int		unpack_hindexed();
static int		unpack_struct();
static int		unpack_char();
static int		unpack_int();
static int		unpack_long();
static int		unpack_short();
static int		unpack_float();
static int		unpack_double();
static int		unpack_long_double();
static int		unpack_complex();
#if __STDC__
static int		unpack_double_complex();
#endif

/*
 * static variables
 */
static int		lowcount;

/*
 *	lam_unpack
 *
 *	Function:	- unpack a message from contiguous space
 *			- unpacking is done recursively
 *			- running out of buffer is an error
 *	Accepts:	- packed buffer
 *			- packed buffer length
 *			- buffer
 *			- count
 *			- datatype
 *	Returns:	- # bytes unpacked or LAMERROR
 */
int
lam_unpack(pkbuf, pklen, buf, count, dtype)

char			*pkbuf;
int			pklen;
char			*buf;
int			count;
MPI_Datatype		dtype;

{
	int		size;			/* # bytes unpacked */
/*
 * Handle the trivial cases.
 */
	if (pklen < 0) return(LAMERROR);
	else if (pklen == 0) return(0);
/*
 * Handle the different datatypes.
 */
	switch(dtype->dt_format) {

	case LAM_DTBASIC:
		if ((dtype == MPI_UNSIGNED_CHAR) || (dtype == MPI_PACKED)
				|| (dtype == MPI_BYTE) || (dtype == MPI_CHAR)
				|| (dtype == MPI_F_CHARACTER)) {
			
			size = unpack_char(pkbuf, pklen, buf, count);
		}
		else if ((dtype == MPI_INT) || (dtype == MPI_UNSIGNED)
				|| (dtype == MPI_F_LOGICAL)
				|| (dtype == MPI_F_INTEGER)) {

			size = unpack_int(pkbuf, pklen, buf, count);
		}
		else if ((dtype == MPI_LONG)
				|| (dtype == MPI_UNSIGNED_LONG)) {

			size = unpack_long(pkbuf, pklen, buf, count);
		}
		else if ((dtype == MPI_SHORT)
				|| (dtype == MPI_UNSIGNED_SHORT)) {

			size = unpack_short(pkbuf, pklen, buf, count);
		}
		else if ((dtype == MPI_FLOAT) || (dtype == MPI_F_REAL)) {

			size = unpack_float(pkbuf, pklen, buf, count);
		}
		else if ((dtype == MPI_DOUBLE)
				|| (dtype == MPI_F_DOUBLE_PRECISION)) {

			size = unpack_double(pkbuf, pklen, buf, count);
		}
		else if (dtype == MPI_LONG_DOUBLE) {
#if __STDC__
			size = unpack_long_double(pkbuf, pklen, buf, count);
#else
			size = unpack_double(pkbuf, pklen, buf, count);
#endif
		}
		else if (dtype == MPI_F_COMPLEX) {

			size = unpack_complex(pkbuf, pklen, buf, count);
		}
		else if (dtype == MPI_F_DOUBLE_COMPLEX) {

			size = unpack_double_complex(pkbuf, pklen, buf, count);
		}
		else if ((dtype == MPI_UB) || (dtype == MPI_LB)) {

			size = 0;
		}
		else {
			size = LAMERROR;
		}

		break;

	case LAM_DTCONTIG:
		size = lam_unpack(pkbuf, pklen, buf,
				count * dtype->dt_count, dtype->dt_dtype);
		break;

	case LAM_DTVECTOR:
		size = unpack_hvector(pkbuf, pklen, buf, count, dtype, 1);
		break;

	case LAM_DTHVECTOR:
		size = unpack_hvector(pkbuf, pklen, buf, count, dtype, 0);
		break;

	case LAM_DTINDEXED:
		size = unpack_hindexed(pkbuf, pklen, buf, count, dtype, 1);
		break;

	case LAM_DTHINDEXED:
		size = unpack_hindexed(pkbuf, pklen, buf, count, dtype, 0);
		break;

	case LAM_DTSTRUCT:
		size = unpack_struct(pkbuf, pklen, buf, count, dtype);
		break;

	default:
		size = LAMERROR;
		break;
	}

	return(size);
}

/*
 *	unpack_hvector
 *
 *	Function:	- unpack vector or hvector
 */
static int
unpack_hvector(frombuf, fromlen, tobuf, num, dtype, fl_vec)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;
MPI_Datatype		dtype;
int			fl_vec;

{
	int		i, j;			/* favourite indices */
	int		size;			/* # bytes unpacked */
	int		totalsize;		/* total # bytes unpacked */
	int		extent;			/* datatype extent */
	int		stride;			/* stride in bytes */
	char		*buf;			/* temp. buffer */

	totalsize = 0;
	extent = dtype->dt_upper - dtype->dt_lower;

	stride = dtype->dt_stride;
	if (fl_vec) stride *= dtype->dt_dtype->dt_upper -
				dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		buf = tobuf;
		tobuf += (tobuf) ? extent : 0;

		for (j = 0; j < dtype->dt_count; ++j) {

			size = lam_unpack(frombuf, fromlen, buf,
					dtype->dt_length, dtype->dt_dtype);

			if (size < 0) return(LAMERROR);

			totalsize += size;
			frombuf += (frombuf) ? size : 0;
			fromlen -= size;
			buf += (buf) ? stride : 0;
		}
	}

	return(totalsize);
}

/*
 *	unpack_hindexed
 *
 *	Function:	- unpack indexed or hindexed
 */
static int
unpack_hindexed(frombuf, fromlen, tobuf, num, dtype, fl_idx)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;
MPI_Datatype		dtype;
int			fl_idx;

{
	int		i, j;			/* favourite indices */
	int		size;			/* # bytes unpacked */
	int		totalsize;		/* total # bytes unpacked */
	int		extent;			/* datatype extent */
	int		factor;			/* extent factor */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	char		*buf;			/* temp. buffer */

	totalsize = 0;
	extent = dtype->dt_upper - dtype->dt_lower;

	factor = (fl_idx == 0) ? 1 :
			dtype->dt_dtype->dt_upper - dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		buf = tobuf;
		tobuf += (tobuf) ? extent : 0;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp) {

			size = lam_unpack(frombuf, fromlen,
						buf + (*disp * factor),
						*len, dtype->dt_dtype);

			if (size < 0) return(LAMERROR);

			totalsize += size;
			frombuf += (frombuf) ? size : 0;
			fromlen -= size;
		}
	}

	return(totalsize);
}

/*
 *	unpack_struct
 *
 *	Function:	- unpack struct
 */
static int
unpack_struct(frombuf, fromlen, tobuf, num, dtype)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;
MPI_Datatype		dtype;

{
	int		i, j;			/* favourite indices */
	int		size;			/* # bytes unpacked */
	int		totalsize;		/* total # bytes unpacked */
	int		extent;			/* datatype extent */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	MPI_Datatype	*type;			/* ptr datatypes */
	char		*buf;			/* temp. buffer */

	totalsize = 0;
	extent = dtype->dt_upper - dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		buf = tobuf;
		tobuf += (tobuf) ? extent : 0;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;
		type = dtype->dt_dtypes;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp, ++type) {

			size = lam_unpack(frombuf, fromlen,
						buf + *disp, *len, *type);

			if (size < 0) return(LAMERROR);

			totalsize += size;
			frombuf += (frombuf) ? size : 0;
			fromlen -= size;
		}
	}

	return(totalsize);
}

/*
 *	unpack_char
 *
 *	Function:	- unpack bytes
 */
static int
unpack_char(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */

	size = (fromlen < num) ? fromlen : num;
	lowcount += size;

	if (tobuf && frombuf) memcpy(tobuf, frombuf, size);

	return(size);
}

/*
 *	unpack_int
 *
 *	Function:	- unpack integers
 */
static int
unpack_int(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # integers */

	n = fromlen / sizeof(int);
	n = (n < num) ? n : num;
	lowcount += n;
	size = n * sizeof(int);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_int(tobuf, n);
	}

	return(size);
}

/*
 *	unpack_long
 *
 *	Function:	- unpack long integers
 */
static int
unpack_long(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # long ints */

	n = fromlen / sizeof(long);
	n = (n < num) ? n : num;
	lowcount += n;
	size = n * sizeof(long);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_long(tobuf, n);
	}

	return(size);
}

/*
 *	unpack_short
 *
 *	Function:	- unpack short integers
 */
static int
unpack_short(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # short ints */

	n = fromlen / sizeof(short);
	n = (n < num) ? n : num;
	lowcount += n;
	size = n * sizeof(short);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_short(tobuf, n);
	}

	return(size);
}

/*
 *	unpack_float
 *
 *	Function:	- unpack single precision floating point numbers
 */
static int
unpack_float(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # floats */

	n = fromlen / sizeof(float);
	n = (n < num) ? n : num;
	lowcount += n;
	size = n * sizeof(float);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_float(tobuf, n);
	}

	return(size);
}

/*
 *	unpack_double
 *
 *	Function:	- unpack double precision floating point numbers
 */
static int
unpack_double(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # doubles */

	n = fromlen / sizeof(double);
	n = (n < num) ? n : num;
	lowcount += n;
	size = n * sizeof(double);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_double(tobuf, n);
	}

	return(size);
}

/*
 *	unpack_complex
 *
 *	Function:	- unpack pairs of floats
 */
static int
unpack_complex(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # floats */

	n = fromlen / (2 * sizeof(float));
	n = (n < num) ? n : num;
	lowcount += n;
	n += n;
	size = n * sizeof(float);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_float(tobuf, n);
	}

	return(size);
}

/*
 *	unpack_double_complex
 *
 *	Function:	- unpack pairs of doubles
 */
static int
unpack_double_complex(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # floats */

	n = fromlen / (2 * sizeof(double));
	n = (n < num) ? n : num;
	lowcount += n;
	n += n;
	size = n * sizeof(double);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_double(tobuf, n);
	}

	return(size);
}

#if __STDC__
/*
 *	unpack_long_double
 *
 *	Function:	- unpack long doubles
 */
static int
unpack_long_double(frombuf, fromlen, tobuf, num)

char			*frombuf;
int			fromlen;
char			*tobuf;
int			num;

{
	int		size;			/* # bytes */
	int		n;			/* # floats */

	n = fromlen / sizeof(long double);
	n = (n < num) ? n : num;
	lowcount += n;
	size = n * sizeof(long double);

	if (tobuf && frombuf) {
		memcpy(tobuf, frombuf, size);
		if (!lam_homog) mttol_longdbl(tobuf, n);
	}

	return(size);
}
#endif

/*
 *	lam_docount
 *
 *	Function:	- get the count of top and bottom level elements
 *	Accepts:	- packed message length
 *			- top level datatype
 *			- ptr top count
 *			- ptr bottom count
 *	Returns:	- unused message length
 */
int
lam_docount(len, dtype, ptop, pbottom)

int			len;
MPI_Datatype		dtype;
int			*ptop;
int			*pbottom;

{
	int		size;			/* # bytes */

	if (dtype->dt_size == 0) {
		*ptop = MPI_UNDEFINED;
		*pbottom = MPI_UNDEFINED;
	}
	else {
		*ptop = ((len % dtype->dt_size) != 0) ?
				MPI_UNDEFINED : (len / dtype->dt_size);

		for (lowcount = 0; len > 0; len -= size) {
			size = lam_unpack((char *) 0, len, (char *) 0,
						1, dtype);
			if (size <= 0) break;
		}

		*pbottom = (len == 0) ? lowcount : MPI_UNDEFINED;
	}
	
	return(len);
}
